// cDataBufferLine.cc -- a class for communicating with MkIII data buffer.
// Copyright (C) 1995 Ari Mujunen. (amn@nfra.nl, Ari.Mujunen@hut.fi)

// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License.
// See the file 'COPYING' for details.

// $Log: cDataBufferLine.cc,v $
// Revision 1.2  1995/07/11  17:21:57  amn
// Memory test diag outputs, split mode.
//
// Revision 1.1  1995/06/16  15:15:16  amn
// Initial revision
//
//

#include "registration.h"
RCSID(cDataBufferLine_cc, "$Id: cDataBufferLine.cc,v 1.2 1995/07/11 17:21:57 amn Exp prog $");


#pragma implementation
#include "cDataBufferLine.hh"

#include "ascii.h"

#include <unistd.h>
#include <assert.h>

// Adjustable behavior details.
#define ALLOW_SET_BLOCK_OPTIMIZATION 1
#define OUR_ERROR_REPORTER fprintf(stderr,
#define DO_ALL_CHECKS 1
#define MAX_GETBLOCK_RETRIES 3
#define WAIT_FOR_HOLDING_RETRIES 10
#define MAX_ERRORS_PER_TEST_BLOCK 5

// Here we transform the meaning of the four additional bytes
// that are included in every response from Mark III data buffer
// (found on pages DB-16..17 of the Data Buffer manual (1-Jan-1982)).
// The sequence is:
//   1) status byte,
//   2) block number 0..255,
//   n-1) id byte, containing a unit number and the memory capacity code,
//   n) the modulo 256 checksum of the characters bytes 1..(n-1).
// And "n" can be either == 4 (the so-called "Identify" response)
// or == 516 (512 bytes of returned data is in between bytes 2 and 515.

// Initially I thought that it would be nice to do this extraction 
// with bit fields.
// Alas, data buffer thinks big endian and Linux/Intel little endian!
// (Even reordering the fields doesn't help, since multibit fields
//  would need opposite bit ordering to retain enum code values the same
//  as in data buffer documentation.)

// So, here are "identify" reply field extraction macros.

#define getParity(x)      (x[0] & 0x80)
#define getOverrun(x)     (x[0] & 0x40)
#define getFraming(x)     (x[0] & 0x20)
#define getMode(x)        ((x[0] & 0x18) >> 3)
#define setTestPattern    (0x04)
#define getTestPattern(x) (x[0] & setTestPattern)
#define getState(x)       (x[0] & 0x03)
typedef enum {
  eStateIdle = 0,
  eStateHolding = 1,
  eStateArmed = 2,
  eStateTaking = 3
} tState;

#define getSetBlockNumber(x) (x[1])

#define getUnit(x)           ((x[2] & 0xf0) >> 4)
#define getMemoryCapacity(x) (x[2] & 0x0f)
typedef enum {
  eMC256kbits = 1,
  eMC512kbits = 2,
  eMC768kbits = 3,
  eMC1024kbits = 4
} tMemoryCapacity;

#define getSetChecksum(x)    (x[3])


// Test patterns built in the data buffer.
#define MAX_TEST_PATTERN (0xc)
typedef unsigned char tSingleTestPattern[4];
typedef unsigned char *ptSingleTestPattern;
static tSingleTestPattern testPattern[MAX_TEST_PATTERN] = {
  {0x00, 0x00, 0x00, 0x00},
  {0xff, 0xff, 0xff, 0xff},
  {0x00, 0x00, 0xff, 0xff},
  {0xff, 0xff, 0x00, 0x00},
  {0x00, 0xff, 0xff, 0x00},
  {0xff, 0x00, 0x00, 0xff},
  {0x55, 0x55, 0xaa, 0xaa},
  {0xaa, 0xaa, 0x55, 0x55},
  {0x55, 0x00, 0x00, 0x55},
  {0x00, 0x55, 0x55, 0x00},
  {0xaa, 0xff, 0xff, 0xaa},
  {0xff, 0xaa, 0xaa, 0xff}
  // Note that the last pattern is a pseudorandom one
  // and cannot be stored for checking.
};


// xxx
static char escAsString[] = {eAsciiEsc, 0};



// A constructor to have initialization of class member fields.
cDataBufferLine::cDataBufferLine(void)
{
  // Well, g++ warned me that ANSI C++ doesn't allow initialization
  // of member fields in class declaration, so I'm having an explicit
  // initialization constructor.

  // Data buffer initial status is expected to be "not helloed".
  status = eDataBufferInactive;

  // Set all "expected" values of "identify" replies to indeterminate,
  // so they won't be reported until initial values are got
  // in 'hello()'.
  currentBlockNumber = -1;
  isExternal = -1;
  currentMode = -1;
  isTestPattern = -1;
  currentlyLoadedTestPattern = -1;
  retainedUnit = -1;
  retainedMemoryCapacity = -1;
}  // cDataBufferLine constructor


// xxx
void
cDataBufferLine::defaultSettings(void)
{
  unsigned char ident[4];
  static char autobaud[] = "auto-baud";
  int i;
  static int baudRate[] = { 115200, 9600, 4800, 2400, 1200, 115200 };
  static int timeOut[] = {       2,    7,   10,   10,   10,      2 };
    // 516 bytes (our longest reply message)
    // takes about 538ms in theory at 9600 baud
    // and about 45ms in theory at 115200 baud.
  
  // Basic serial initialization via the inherited 'defaultSettings()'.
  cSerialLine::defaultSettings();

  for (i=0; i<(sizeof(baudRate)/sizeof(int)); i++) {
    OUR_ERROR_REPORTER "%s: trying %d baud...",
                       autobaud,
                       baudRate[i]);
    serialSettings(baudRate[i],
		   8,
		   1,
		   eParityNone,
		   0,  // minBytes; the shortest message we expect
		       // (excluding no answer ie. timeout)
		   timeOut[i]  // maxTimeout in 0.1 sec units
		   );

    // If the data buffer responds to hello/status query with 4 bytes,
    // assume ok.
    (void)justSendString(escAsString, sizeof(ident), ident, autobaud);
    if (status == eDataBufferOk) {
      // Undo the successful 'hello' (esc) with 'goodbye()'.
      goodbye();
      OUR_ERROR_REPORTER "ok, using %d baud.\n",
                         baudRate[i]);
      return;
    } 
  }  // for each baud rate

  // Too bad, no baud rate could coax a 4-byte reply from data buffer.
  OUR_ERROR_REPORTER "%s: data buffer didn't respond with any baud rate tried.\n",
                       autobaud);
  // xxx: status = ?

  // xxx: Linux 1.2.8 'serial.c' seems to wait forever for minBytes>0,
  // even when VTIME>0; this should have applied to each
  // new character received, not to the whole message.
  // Must test again in the future---we always expect at least 4 bytes back.

  // xxx: Eventually might try to add auto detection of
  //      hw handshake enabling cable (perhaps to cSerialLine).
}  // defaultSettings


// Send the readily-formatted DataBuffer message string 's'
// and receive the reply of length 'expectedReplyLength'.
// Only report timeout error, no others.
int
cDataBufferLine::justSendString(const char *s,
				int expectedReplyLength,
				unsigned char *reply,
				const char *command
				)
{
  flushPendingInput();
  write(s, strlen(s));
  status = eDataBufferOk;
  if (expectedReplyLength <= 0) {
    return (0);  // no errors
  }

  // xxx
  if ((lengthOfLatestReply
       = read(reply, expectedReplyLength)) < expectedReplyLength) {
    status = eDataBufferTimeout;
    OUR_ERROR_REPORTER "%s: a reply of only %d bytes (%d expected)\n",
                       command,
		       lengthOfLatestReply,
		       expectedReplyLength
                       );
    return (1);  // one error
  }
  return (0);  // no errors
}  // justSendString

// Send the readily-formatted DataBuffer message string 's'
// and receive the reply of length 'expectedReplyLength'.
// Report the number of errors encountered in reply.
int
cDataBufferLine::sendString(const char *s,
		            int expectedReplyLength,
		            unsigned char *reply,
			    const char *command
		            )
{
  int retVal;
  
  if ((retVal = justSendString(s, expectedReplyLength,reply, command))) {
    return retVal;
  }
  
  // Long (516 byte) replies and those short replies that
  // go to somewhere else than our own instance 'identify' field
  // need copying.
  if ((expectedReplyLength > 4) || ((reply != identify))) {
    // We need to plant the 2 prefix and 2 postfix bytes
    // into our member 'identify' field.
    identify[0] = reply[0];
    identify[1] = reply[1];
    identify[2] = reply[lengthOfLatestReply-2];
    identify[3] = reply[lengthOfLatestReply-1];
  } else if (expectedReplyLength == 4) {
    // We are doing a "dangerous" optimization by expecting this
    // to be always to our own member 'identify' field.
    assert(((expectedReplyLength == 4), (reply == identify)));
    // So nothing needs to be assigned to 'identify'.
  } else {
    assert(expectedReplyLength >= 4);  // will always fail
  }

#if 0
    OUR_ERROR_REPORTER "%s: identify= %02X %02X %02X %02X\n",
                       command,
		       identify[0], identify[1], identify[2], identify[3]
                       );
#endif

  // Calculate checksum.
  {
    int newSum = 0;
    int i;

    for (i=0; i<(lengthOfLatestReply-1); i++) {
      newSum += reply[i];
    }
    newSum &= 0xff;
    return (generalStatusCheck(newSum, command));
  }  // block for checksum
}  // sendString


#define ENSURE_ACTIVE(op) \
  if (status == eDataBufferInactive) { \
    OUR_ERROR_REPORTER "operation '%s' to an inactive data buffer\n", op); \
    OUR_ERROR_REPORTER "(performing an automatic 'hello(internal)' to it)\n"); \
    hello(eDataSourceInternal); \
  }

#define ENSURE_INACTIVE(op) \
  if (status != eDataBufferInactive) { \
    OUR_ERROR_REPORTER "operation '%s' to an already active data buffer\n", op); \
    OUR_ERROR_REPORTER "(performing an automatic 'goodbye()' to it)\n"); \
    goodbye(); \
  }


#define COMPLAIN(bit, what) \
  if ((bit)) { \
    errors++; \
    OUR_ERROR_REPORTER "a%s error sending command '%s'\n", \
                       what, command); \
  }

#define SHOULD_BE(old, new, what) \
{ \
  int oldi = (old); \
  int newi = (new); \
 \
  if ((oldi >= 0) && (oldi != newi)) { \
    errors++; \
    OUR_ERROR_REPORTER "%s: '%s' should be %d and it is %d\n", \
                       command, what, oldi, newi); \
  } \
}

#define SHOULD_NOT_BE(old, new, what) \
{ \
  int oldi = (old); \
  int newi = (new); \
 \
  if ((oldi >= 0) && (oldi == newi)) { \
    errors++; \
    OUR_ERROR_REPORTER "%s: '%s' should NOT be %d and it is\n", \
                       command, what, newi); \
  } \
}

// xxx
int
cDataBufferLine::generalStatusCheck(int newChecksum, const char *command)
{
  int errors = 0;

#if DO_ALL_CHECKS
  SHOULD_BE(newChecksum, getSetChecksum(identify),
    "checksum");

  // We are not going to process a garbled message.
  if (newChecksum != getSetChecksum(identify)) {
    status = eDataBufferChecksum;
    return (errors);
  }

  // These can be checked for every response.
  COMPLAIN(getParity(identify), " parity");
  COMPLAIN(getOverrun(identify), "n overrun");
  COMPLAIN(getFraming(identify), " framing");
  state = getState(identify);  // checked after 'general...' is done

  // These can be checked after 'hello()'.
  SHOULD_BE(currentMode, getMode(identify),
    "data buffer mode");
  //SHOULD_BE(isExternal, getExternal(identify), "old" data buff.
  //  "using external signal");
  SHOULD_BE(retainedUnit, getUnit(identify),
    "data buffer unit id number");
  SHOULD_BE(retainedMemoryCapacity, getMemoryCapacity(identify),
    "memory capacity code");

  // These change in the course of actions on data buffer.
  SHOULD_BE(isTestPattern, getTestPattern(identify),
    "buffer contains test pattern");
  SHOULD_BE(currentBlockNumber, getSetBlockNumber(identify),
    "received block number");
#endif

  return (errors);
}  // generalStatusCheck


#define DECLARE_COMMAND(s) \
  static char command[] = s; \
  int errors = 0


// xxx
void
cDataBufferLine::hello(tDataSource which)
{
  DECLARE_COMMAND("hello");

  ENSURE_INACTIVE(command);
  errors += sendString(escAsString, sizeof(identify), identify, command);
  // xxx: this is the only place we cannot do 'general' inside 'sendString'
  // perhaps do a couple of times,
  // do we get the same identify string every time?

  // Save "should_be" numbers for future checks.
  //isExternal = getExternal(identify); old data buffer
  currentMode = getMode(identify);
  retainedUnit = getUnit(identify);
  retainedMemoryCapacity = getMemoryCapacity(identify);
  maxBlockNumberSupportedByThisMemorySize = (64 * retainedMemoryCapacity) - 1;

  // Complain if using wrong data source, state switch position
  // in data buffer rear panel.
#if 0
old
  if ( (isExternal == 0) != (which == eDataSourceInternal) ) {
#else
  if (currentMode == eModeExternal) {
#endif
    // xxx: check
    OUR_ERROR_REPORTER "%s: 'INT/EXT' switch at data buffer rear panel is in wrong '%s' position; this software expects it to be in '%s' position, please adjust\n", command, (isExternal?"EXT":"INT"), ((which == eDataSourceInternal)?"INT":"EXT"));
  }

  // Just to make sure, do a reset.
  // xxx: perhaps when != idle only?
  reset();
  setNextBlock(0);  // perhaps?
}  // hello

// xxx
void
cDataBufferLine::goodbye(void)
{
  DECLARE_COMMAND("goodbye");

  ENSURE_ACTIVE(command);
  errors += sendString("%", sizeof(identify), identify, command);
  // xxx: checkl success?
  // xxx: In principle we could '-1' all the expected values,
  //      but in practice the data buffer retains them while inactive.
  status = eDataBufferInactive;
}  // goodbye

// xxx
void
cDataBufferLine::reset(void)
{
  DECLARE_COMMAND("reset");

  ENSURE_ACTIVE(command);
  errors += sendString(":", sizeof(identify), identify, command);

  // Verify that data buffer goes idle.
  SHOULD_BE(eStateIdle, state,
    "state afterwards");
}  // reset

// xxx
void
cDataBufferLine::setNextBlock(int aBlockNumber)
{
  DECLARE_COMMAND("setNextBlock");
  char temp[4];  // "nn+\0"
  int savedBlockNumber;

  ENSURE_ACTIVE(command);
  if ((aBlockNumber < 0)
        || (aBlockNumber > maxBlockNumberSupportedByThisMemorySize)) {
    // xxx: status = something "didn't do it"
    OUR_ERROR_REPORTER "%s: requested block %d, memory size in data buffer supports only 0..%d", command, aBlockNumber, maxBlockNumberSupportedByThisMemorySize);
    return;
  }

  // And just to make sure only two hex chars will appear:
  aBlockNumber &= 0xff;
  assert(sprintf(temp, "%02X+", aBlockNumber) < sizeof(temp));

  // Change expected block number in advance, saving the old
  // one for the case change fails.
  savedBlockNumber = currentBlockNumber;
  currentBlockNumber = aBlockNumber;

  errors += sendString(temp, sizeof(identify), identify, command);

  // If the state of data buffer was holding or idle,
  // the block number was set---otherwise (armed/taking) it had no effect.
  if ( (state == eStateIdle) || (state == eStateHolding) ) {
    // currentBlockNumber is already set
  } else {
    // Restore the block number to what there already is in data buffer.
    currentBlockNumber = savedBlockNumber;
      // xxx: or get from identify response?
    // xxx: status = something
    OUR_ERROR_REPORTER "%s: cannot set active block while state=%d\n",
                       command,
                       state);
  }
}  // setNextBlock

// xxx
void
cDataBufferLine::arm(void)
{
  DECLARE_COMMAND("arm");

  ENSURE_ACTIVE(command);

  // Verify that data buffer is idle (page DB-14 (1-Jan-1982) implies
  // explicitly that you cannot arm while 'holding'.)
  SHOULD_BE(eStateIdle, state,
    "state before arming");
  if (state != eStateIdle) {
    OUR_ERROR_REPORTER "%s: (performing an automatic 'reset()')\n", command);
    reset();
  }

  // Next communication should return 'test pattern on' bit == 0,
  // so we change our expectations now:
  isTestPattern = 0;
  // And since the test pattern will be gone, forget its number:
  currentlyLoadedTestPattern = -1;
  // And the returned block number will be indeterminate until
  // actually holding.
  currentBlockNumber = -1;

  errors += sendString("!", sizeof(identify), identify, command);

  // Check state armed (-> taking -> holding), ie. NOT idle anymore.
  SHOULD_NOT_BE(eStateIdle, state,
    "state after arming");
  if (state == eStateIdle) {
    // xxx: status = something "didn't do it/hw failed"
  }
}  // arm

// xxx
void
cDataBufferLine::sendData(tBlock aBlock)
{
  DECLARE_COMMAND("sendData");
  static unsigned char blockReply[516];

  ENSURE_ACTIVE(command);
  // xxx: needs to be holding (or idle), check, reset?
  // xxx: 'idle' ok if especially allowed: garbage from buffer...

  errors += sendString("?", sizeof(blockReply), (unsigned char *)&blockReply, command);
  // xxx: check state+status? timeouts, too short replies?

  // The data buffer counter hardware increments automatically to next block.
  currentBlockNumber++;

  // Checks ok, now lets copy it back.
#if DEBUG
  printf("sendData: currentMode=%d\n", currentMode);
#endif
  if (currentMode == eModeSplit) {
    int i, j;
    const int numOfPairs = sizeof(tBlock)/2;
    
    // Reshuffle the bytes, initially to the same buffer,
    // xxx: perhaps eventually into two buffers.
    for (i=0,j=0; i < numOfPairs; i++,j+=2) {
      aBlock[i] = blockReply[j+2];
      aBlock[i+numOfPairs] = blockReply[j+3];
    }  // for each pair of bytes
  } else {
    // Just block move the 512 bytes.
    memcpy(aBlock, &(blockReply[2]), sizeof(tBlock));
  }
}  // sendData

// xxx
void
cDataBufferLine::setMode(int newMode)
{
  DECLARE_COMMAND("setMode");
  char temp[3];  // "n&\0"

  ENSURE_ACTIVE(command);
  if ((newMode < 0) || (newMode > 3)) {
    // xxx: status = something "didn't do it"
    OUR_ERROR_REPORTER "%s: requested mode %d, data buffer supports only 0..%d", command, newMode, 3);
    return;
  }

  // Verify that data buffer is idle.
  SHOULD_BE(eStateIdle, state,
    "state before changing mode");
  if (state != eStateIdle) {
    OUR_ERROR_REPORTER "%s: (performing an automatic 'reset()')\n", command);
    reset();
  }

  // Next communication should return the new mode bits
  // so we change our expectations now:
  currentMode = (newMode & 0x3);

  assert(sprintf(temp, "%01X&", currentMode) < sizeof(temp));
  errors += sendString(temp, sizeof(identify), identify, command);
  // xxx: success?
}  // setMode

// xxx
void
cDataBufferLine::startTest(int testNumber)
{
  DECLARE_COMMAND("startTest");
  char temp[3];  // "n-\0"

  ENSURE_ACTIVE(command);
  if ((testNumber < 0)
        || (testNumber > MAX_TEST_PATTERN)) {
    // xxx: status = something "didn't do it"
    OUR_ERROR_REPORTER "%s: requested test pattern %d, data buffer supports only 0..%d", command, testNumber, MAX_TEST_PATTERN);
    return;
  }

  // Verify that data buffer is idle.
  SHOULD_BE(eStateIdle, state,
    "state before establishing a test pattern");
  if (state != eStateIdle) {
    OUR_ERROR_REPORTER "%s: (performing an automatic 'reset()')\n", command);
    reset();
  }

  // Next communication should return 'test pattern on' bit == 1,
  // so we change our expectations now:
  isTestPattern = setTestPattern;
  // And the returned block number will be indeterminate until
  // actually holding.
  currentBlockNumber = -1;

  assert(sprintf(temp, "%01X-", (testNumber & 0xf)) < sizeof(temp));
  errors += sendString(temp, sizeof(identify), identify, command);

  if ( !((state == eStateTaking) || (state == eStateHolding)) ) {
    // xxx: status = something "hw didn't obey"
    OUR_ERROR_REPORTER "%s: data buffer didn't start loading test pattern %d, state=%d\n",
                       command,
                       testNumber,
                       state);
    // No test pattern after all:
    isTestPattern = 0;
    // xxx: might want to reset the block number?
    return;
  }

  // Hardware started loading the pattern, let's remember the pattern number.
  currentlyLoadedTestPattern = testNumber;
}  // startTest


// Combined actions.
// xxx
void
cDataBufferLine::askStatus(void)
{
  DECLARE_COMMAND("askStatus");

  ENSURE_ACTIVE(command);
  errors += sendString(escAsString, sizeof(identify), identify, command);
  // prepare status?
}  // askStatus

// waitforHolding() -- this can fail == timeout -> reset -> idle, not holding
// Timeout is in 0.01sec (100Hz) units.
void
cDataBufferLine::waitForHolding(int timeout)
{
  int i;

  // We try max WAIT_FOR_HOLDING_RETRIES times,
  // at an interval (timeout/WAIT_FOR_HOLDING_RETRIES)
  for (i=0; i<WAIT_FOR_HOLDING_RETRIES; i++) {
    askStatus();
    if (state == eStateHolding) {
      // 'status' is ok, just return:
      return;
    }
    usleep(10000*timeout/WAIT_FOR_HOLDING_RETRIES);
  }
  OUR_ERROR_REPORTER "waitForHolding: timeout, state 'holding' not reached within %g seconds)\n", (double)timeout/100.0);
  reset();
  // xxx: status = holding-timeout
}  // waitForHolding

// xxx
void
cDataBufferLine::armAndLoad(int timeout)
{
  arm();
  waitForHolding(timeout);
}  // armAndLoad

// xxx
void
cDataBufferLine::loadTest(int testNumber, int timeout)
{
  startTest(testNumber);
  waitForHolding(timeout);
}  // loadTest

// xxx
void
cDataBufferLine::getBlock(int aBlockNumber, tBlock aBlock)
{
  int retries = MAX_GETBLOCK_RETRIES;

  do {
    if ((currentBlockNumber == aBlockNumber) && ALLOW_SET_BLOCK_OPTIMIZATION) {
      ;  // do nothing, because the data buffer should be at this block
    } else {
      setNextBlock(aBlockNumber);
    }
    sendData(aBlock);
    retries--;
  } while (retries && (status != eDataBufferOk));
}  // getBlock


static void
bits(FILE *toWhere,
     const char *s,
     int byte)
{
  int i;
  
  for (i=0; i<8; i++) {
    if ((byte >> i) & 0x01) {
      fprintf(toWhere, s, i);
    }
  }  // for
}  // bits

// xxx
void
cDataBufferLine::checkTestBlock(int aBlockNumber,
				tCheckInfo& ci,
				int suppressMessages,
				FILE *toWhere)
{
  if ( (currentlyLoadedTestPattern < 0)
      || (currentlyLoadedTestPattern > MAX_TEST_PATTERN - 1) ) {
    // xxx: status = "sw logic error"
    fprintf(toWhere, "checkTestBlock: cannot verify current test pattern number %d (outside 0..%d)\n",
                       currentlyLoadedTestPattern,
                       MAX_TEST_PATTERN - 1);
    return;
  }

  // Now we know that the data buffer is loaded with a pattern
  // which we have tabulated and can thus compare.
  {
    tBlock aBlock;
    ptSingleTestPattern tp;
    int i;
    int patternByte;
    int patternIndex;
    int errors;

    getBlock(aBlockNumber, aBlock);

    tp = testPattern[currentlyLoadedTestPattern];

    // First swoop thru to count erroneous bytes
    // and accumulate bit patterns that are erroneous.
    ci.clearCheckInfo();
    for (i=0; i<sizeof(aBlock); i++) {
      if (aBlock[i] != (patternByte = tp[(patternIndex = i % 4)])) {
	ci.errors++;
	ci.maintainZeroAddressBits |= i;
	ci.maintainOneAddressBits |= ~i;
	ci.zeroesSeen[patternIndex] |= ~aBlock[i];
	ci.onesSeen[patternIndex] |= aBlock[i];
	ci.stuckToZero[patternIndex] |= ((~aBlock[i]) & patternByte);
	ci.stuckToOne[patternIndex] |= (aBlock[i] & (~patternByte));
	ci.errorsPerIndex[patternIndex]++;
      }
    }  // for each byte

    // Rest of this function deals with printing out what's wrong.
    if (suppressMessages) {
      return;
    }

    // Start considering printing out error messages.
    // Perhaps in condensed format for repeated errors?
    if (ci.errors >= (512/4)) {  // repeats at least once per 4 pattern bytes
      int whichAreStuck;
      int surelyStuckToZero[4];
      int surelyStuckToOne[4];

      fprintf(toWhere,
	      "checkTestBlock: with pattern=%d, block=%d (in memory bank %d):\n  suspecting a stuck bit problem\n  because there are %d (==%d or more) errors per a 512-byte block:\n",
	      currentlyLoadedTestPattern,
	      aBlockNumber,
	      (aBlockNumber >> 6),
	      ci.errors,
	      (512/4)
	      );
      fprintf(toWhere,
	      "lsb of address/hi-lo-chips:  0/lo  1/hi  2/lo  3/hi\n");
      fprintf(toWhere,
	      "number of errors:             %3d   %3d   %3d   %3d,  total errors = %d\n",
	      ci.errorsPerIndex[0],
	      ci.errorsPerIndex[1],
	      ci.errorsPerIndex[2],
	      ci.errorsPerIndex[3],
	      ci.errors
	      );
      fprintf(toWhere,
	      "test pattern 0x%01X:            0x%02X  0x%02X  0x%02X  0x%02X\n",
	      currentlyLoadedTestPattern,
	      (int)tp[0],
	      (int)tp[1],
	      (int)tp[2],
	      (int)tp[3]
	      );

      whichAreStuck =
	ci.stuckToZero[0]
	  | ci.stuckToZero[1]
	    | ci.stuckToZero[2]
	      | ci.stuckToZero[3];
      if (whichAreStuck) {
	fprintf(toWhere,
		"bits possibly stuck to zero: 0x%02X  0x%02X  0x%02X  0x%02X\n",
		ci.stuckToZero[0],
		ci.stuckToZero[1],
		ci.stuckToZero[2],
		ci.stuckToZero[3]
		);
	// Eliminate the bits that have exhibited occasional ones.
	whichAreStuck = (surelyStuckToZero[0] = ci.stuckToZero[0] & ~ci.onesSeen[0]);
	whichAreStuck |= (surelyStuckToZero[1] = ci.stuckToZero[1] & ~ci.onesSeen[1]);
	whichAreStuck |= (surelyStuckToZero[2] = ci.stuckToZero[2] & ~ci.onesSeen[2]);
	whichAreStuck |= (surelyStuckToZero[3] = ci.stuckToZero[3] & ~ci.onesSeen[3]);

	if (whichAreStuck) {
	  fprintf(toWhere,
		  "bits probably stuck to zero: 0x%02X  0x%02X  0x%02X  0x%02X,  together 0x%02X\n",
		  surelyStuckToZero[0],
		  surelyStuckToZero[1],
		  surelyStuckToZero[2],
		  surelyStuckToZero[3],
		  whichAreStuck
		  );
	  {
	    int stuckToZeroHi, stuckToZeroLo;
	    
	    // Determine whether the stuck bit is more likely
	    // in 16-bit memory in path or 8-bit memory out path.
	    stuckToZeroHi = surelyStuckToZero[1] | surelyStuckToZero[3];
	    stuckToZeroLo = surelyStuckToZero[0] | surelyStuckToZero[2];
	    if (stuckToZeroHi != stuckToZeroLo) {
	      fprintf(toWhere, "  Suggesting HD/LD 16-bit data path from shift registers to memory\n  since hi and lo byte are stuck in a different manner\n  (see schematics 6004-008..-012):\n");
	      bits(toWhere, "  HD%d stuck to zero\n", stuckToZeroHi);
	      bits(toWhere, "  LD%d stuck to zero\n", stuckToZeroLo);
	    } else {
	      fprintf(toWhere, "  suggesting MD 8-bit data path from memory to UART\n  since both hi and lo byte are stuck in the same way\n  (see schematics 6004-009..-012 (bus MD) and -004, -005 (buses MD and TBR):");
	      bits(toWhere, "  MD%d stuck to zero\n", stuckToZeroHi);
	    }  // else MD bus
	  }  // block for bus guess
	}  // if any bit probably stuck
      }  // if any bits possibly stuck

      whichAreStuck =
	ci.stuckToOne[0]
	  | ci.stuckToOne[1]
	    | ci.stuckToOne[2]
	      | ci.stuckToOne[3];
      if (whichAreStuck) {
	fprintf(toWhere,
		"bits possibly stuck to one:  0x%02X  0x%02X  0x%02X  0x%02X\n",
		ci.stuckToOne[0],
		ci.stuckToOne[1],
		ci.stuckToOne[2],
		ci.stuckToOne[3]
		);
	// Eliminate the bits that have exhibited occasional zeroes.
	whichAreStuck = (surelyStuckToOne[0] = ci.stuckToOne[0] & ~ci.zeroesSeen[0]);
	whichAreStuck |= (surelyStuckToOne[1] = ci.stuckToOne[1] & ~ci.zeroesSeen[1]);
	whichAreStuck |= (surelyStuckToOne[2] = ci.stuckToOne[2] & ~ci.zeroesSeen[2]);
	whichAreStuck |= (surelyStuckToOne[3] = ci.stuckToOne[3] & ~ci.zeroesSeen[3]);
	if (whichAreStuck) {
	  fprintf(toWhere,
		  "bits probably stuck to one:  0x%02X  0x%02X  0x%02X  0x%02X,  together 0x%02X\n",
		  surelyStuckToOne[0],
		  surelyStuckToOne[1],
		  surelyStuckToOne[2],
		  surelyStuckToOne[3],
		  whichAreStuck
		  );
	  {
	    int stuckToOneHi, stuckToOneLo;
	    
	    // Determine whether the stuck bit is more likely
	    // in 16-bit memory in path or 8-bit memory out path.
	    stuckToOneHi = surelyStuckToOne[1] | surelyStuckToOne[3];
	    stuckToOneLo = surelyStuckToOne[0] | surelyStuckToOne[2];
	    if (stuckToOneHi != stuckToOneLo) {
	      fprintf(toWhere, "  Suggesting HD/LD 16-bit data path from shift registers to memory\n  since hi and lo byte are stuck in a different manner\n  (see schematics 6004-008..-012):\n");
	      bits(toWhere, "    HD%d stuck to one\n", stuckToOneHi);
	      bits(toWhere, "    LD%d stuck to one\n", stuckToOneLo);
	    } else {
	      fprintf(toWhere, "  suggesting MD 8-bit data path from memory to UART\n  since both hi and lo byte are stuck in the same way\n  (see schematics 6004-009..-012 (bus MD) and -004, -005 (buses MD and TBR):");
	      bits(toWhere, "    MD%d stuck to one\n", stuckToOneHi);
	    } 
	  }  // block for MD/HDLD bus detection
	}  // if any bit is probably stuck
      }  // if any bit is possibly stuck

      // No detailed byte-by-byte report.
      return;
    }

    // Not too many errors found, scan again to print out
    // the occasional ones.
    errors = 0;
    for (i=0; i<sizeof(aBlock); i++) {
      if (aBlock[i] != tp[i % 4]) {
	if (errors++ > MAX_ERRORS_PER_TEST_BLOCK) {
	  fprintf(toWhere,
		  "checkTestBlock: with pattern=%d, block=%d (in memory bank %d):\n  more than %d errors, skipping the rest of the block...\n",
		  currentlyLoadedTestPattern,
		  aBlockNumber,
		  (aBlockNumber >> 6),
		  MAX_ERRORS_PER_TEST_BLOCK);
	  break;
	}
        fprintf(toWhere,
		"checkTestBlock: with pattern=%d, block=%d (in memory bank %d):\n  byte=%d should be 0x%02X and it is 0x%02X\n",
		currentlyLoadedTestPattern,
		aBlockNumber,
		(aBlockNumber >> 6),
		i,
		tp[i % 4],
		aBlock[i]
		);
      }
    }  // for each byte
  }  // block
}  // checkTestBlock


// A higher-level test procedure.

/*
hello(source)
goodbye
hello
reset
arm
 - armed(/taking/holding)
reset (as soon as >=armed)
 - idle
startTest
 - taking(/holding)
reset (as soon as >= taking)
 - idle
loadTest(0..11?)
 - check 1 block
loadTest(0..11?)
 - check all blocks

xxx: add external reset via decoder: int/ext switch problems
xxx: to decoder: test error counters (perhaps dead bits, too)
(three counter banks in hw: errors, data, time tag)

xxx: remember to try 18MHz at ext clock input + test patterns
*/
