// cDataBufferLine.hh -- a class for communicating with MkIII data buffer.
// Copyright (C) 1995 Ari Mujunen. (amn@nfra.nl, Ari.Mujunen@hut.fi)

// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License.
// See the file 'COPYING' for details.

// $Log: cDataBufferLine.hh,v $
// Revision 1.2  1995/07/11  17:22:22  amn
// Split mode.
//
// Revision 1.1  1995/06/16  15:18:11  amn
// Initial revision
//
//


#ifndef _cDataBufferLine_hh
#ifdef __GNUG__
#pragma interface
#endif
#define _cDataBufferLine_hh 1

#include "registration.h"
RCSID(cDataBufferLine_hh, "$Id: cDataBufferLine.hh,v 1.2 1995/07/11 17:22:22 amn Exp $");


// 'cDataBufferLine' inherits the basic behavior of a serial interface
// from 'cSerialLine', so we need its header.
#include "cSerialLine.hh"

// For 'FILE *'.
#include <stdio.h>
// For 'memset()'.
#include <string.h>


// Possible responses to the 'getStatus()' query.
typedef enum {eDataBufferOk,  // all went well
	      eDataBufferTimeout,  // device is not there or it is dead
	      eDataBufferChecksum,  // reply had a checksum error, ignored
	      eDataBufferInactive,  // has not been woken up with ESC
	      // xxx: more responses?
	      } tDataBuffer;

// For calls that return data blocks from the buffer
typedef unsigned char tBlock[512];

// For specifying in which position we expect the data source
// switch to be when we say 'hello()' to the data buffer.
typedef enum {eDataSourceInternal, eDataSourceExternal} tDataSource;

// For specifying data buffer operating mode (input handling).
typedef enum {eModeDecoderA = 0x0,
	      eModeExternal,
	      eModeDecoderB,
	      eModeSplit
	      } tMode;

// For storing and comparing error results of 'checkTestBlock()'.
typedef class cCheckInfo {
public:
  int errors;
  int maintainZeroAddressBits;
  int maintainOneAddressBits;
  int zeroesSeen[4];
  int onesSeen[4];
  int stuckToZero[4];
  int stuckToOne[4];
  int errorsPerIndex[4];
  void clearCheckInfo(void) {
    memset(this, 0, sizeof(tCheckInfo));
  }
  int operator ==(const class cCheckInfo& x) {
    return (!memcmp(this, &x, sizeof(class cCheckInfo)));
  };
  int operator !=(const class cCheckInfo& x) {
    return (!(*this == x));
  };
} tCheckInfo, *ptCheckInfo;


// This class provides the low-level means of communicating
// with DataBuffer serial devices.
class cDataBufferLine : public cSerialLine
// xxx: We might opt for inheriting in 'protected' mode
// to prevent some of the 'cSerialLine' methods being visible.
{
public:
  // A constructor to have initialization of class member fields.
  cDataBufferLine(void);
  virtual void defaultSettings(void);
  virtual tDataBuffer getStatus(void) { return (status); };
private:
  // The length of latest reply, ok cases are 4-byte "Identify"
  // and 516-byte "Data Block" messages.
  int lengthOfLatestReply;

  // For the incoming four-byte "Identify" messages:
  unsigned char identify[4];

  // Decoded from the previous, one of the hardware states 0..3,
  // see implementation.
  int state;

  // The block number we expect to be in data buffer
  // block number counter right now.
  int currentBlockNumber;

  // Other "should be" numbers for status checks.
  int isExternal;
  int currentMode;
  int isTestPattern;
  int currentlyLoadedTestPattern;  // or -1 for none
  int retainedUnit;
  int retainedMemoryCapacity;
  // Converted from memory capacity:
  int maxBlockNumberSupportedByThisMemorySize;

protected:
  tDataBuffer status;

  // xxx
  virtual int justSendString(const char *s,
			     int expectedReplyLength,
			     unsigned char *reply,
			     const char *command
			     );
  virtual int sendString(const char *s,
			 int expectedReplyLength,
			 unsigned char *reply,
			 const char *command
			 );

  // Returns the number of errors found in the "routine" status check.
  virtual int generalStatusCheck(int newChecksum, const char *command);

public:
  // The data buffer "primitives" from page DB-14 of Mark III manual.
  virtual void hello(tDataSource which);
  virtual void goodbye(void);
  virtual void reset(void);
  virtual void setNextBlock(int aBlockNumber);
  virtual void arm(void);
  virtual void sendData(tBlock aBlock);
  virtual void setMode(int newMode);
  virtual void startTest(int testNumber);

  // Combined actions.
  virtual void askStatus(void);
  virtual void getBlock(int aBlockNumber, tBlock aBlock);
  virtual void waitForHolding(int timeout);
  virtual void armAndLoad(int timeout);
  virtual void loadTest(int testNumber, int timeout);
  virtual void checkTestBlock(int aBlockNumber,
			      tCheckInfo& ci,
			      int suppressMessages,
			      FILE *toWhere);
};  // class cDataBufferLine

#endif
